mod error;
pub mod models;

#[cfg(test)]
mod tests;

pub use crate::error::{ApiError, ApiErrorKind};

use failure::ResultExt;
use models::{
    Feed, FeedCreateInput, FeedMoveInput, FeedRenameInput, FeedsResponse, Folder, FolderName,
    FoldersResponse, Item, ItemMarkAllInput, ItemMarkInput, ItemQueryInput, ItemType,
    ItemUpdatedQueryInput, ItemsResponse, Version,
};
use reqwest::{header::AUTHORIZATION, Client, StatusCode};
use url::Url;

type FolderID = i64;
type FeedID = i64;
type ItemID = i64;

pub struct NextcloudNewsApi {
    base_uri: Url,
    auth: String,
}

impl NextcloudNewsApi {
    /// Create a new instance of the NextcloudNewsApi.
    /// - `url`: url of the hosted Nextcloud instance (e.g. `https://cloud.test.com/`)
    /// - `username`: user existing on the Nextcloud instance
    /// - `password`: password of the user
    pub fn new(url: &Url, username: String, password: String) -> Self {
        Self {
            base_uri: url.clone(),
            auth: Self::generate_basic_auth(&username, &password),
        }
    }

    async fn parse_error(
        response: reqwest::Response,
        expected_http: StatusCode,
    ) -> Result<String, ApiError> {
        let status = response.status();
        let response = response.text().await.context(ApiErrorKind::Http)?;
        if status != expected_http {
            return Err(ApiErrorKind::Unknown.into());
        }
        Ok(response)
    }

    fn get_api_1_2_uri(&self) -> Result<Url, ApiError> {
        Ok(self
            .base_uri
            .join("index.php/apps/news/api/v1-2/")
            .context(ApiErrorKind::Url)?)
    }

    fn get_api_1_3_uri(&self) -> Result<Url, ApiError> {
        Ok(self
            .base_uri
            .join("index.php/apps/news/api/v1-3/")
            .context(ApiErrorKind::Url)?)
    }

    fn generate_basic_auth(username: &str, password: &str) -> String {
        let auth = format!("{}:{}", username, password);
        let auth = base64::encode(&auth);
        format!("Basic {}", auth)
    }

    pub async fn get_version(&self, client: &Client) -> Result<Version, ApiError> {
        let version = self.get_version_1_3(client).await;

        if let Ok(version) = version {
            Ok(version)
        } else {
            self.get_version_1_2(client).await
        }
    }

    async fn get_version_1_3(&self, client: &Client) -> Result<Version, ApiError> {
        let api_url = self
            .get_api_1_3_uri()?
            .join("version")
            .context(ApiErrorKind::Url)?;

        let response = client
            .get(api_url)
            .header(AUTHORIZATION, self.auth.clone())
            .send()
            .await
            .context(ApiErrorKind::Http)?;

        let response = Self::parse_error(response, StatusCode::OK).await?;
        let version: Version = serde_json::from_str(&response).context(ApiErrorKind::Json)?;
        Ok(version)
    }

    async fn get_version_1_2(&self, client: &Client) -> Result<Version, ApiError> {
        let api_url = self
            .get_api_1_2_uri()?
            .join("version")
            .context(ApiErrorKind::Url)?;

        let response = client
            .get(api_url)
            .header(AUTHORIZATION, self.auth.clone())
            .send()
            .await
            .context(ApiErrorKind::Http)?;

        let response = Self::parse_error(response, StatusCode::OK).await?;
        let version: Version = serde_json::from_str(&response).context(ApiErrorKind::Json)?;
        Ok(version)
    }

    pub async fn get_folders(&self, client: &Client) -> Result<Vec<Folder>, ApiError> {
        let api_url = self
            .get_api_1_3_uri()?
            .join("folders")
            .context(ApiErrorKind::Url)?;

        let response = client
            .get(api_url)
            .header(AUTHORIZATION, self.auth.clone())
            .send()
            .await
            .context(ApiErrorKind::Http)?;

        let response = Self::parse_error(response, StatusCode::OK).await?;
        let response: FoldersResponse =
            serde_json::from_str(&response).context(ApiErrorKind::Json)?;
        let FoldersResponse { folders } = response;
        Ok(folders)
    }

    pub async fn create_folder(&self, client: &Client, name: &str) -> Result<Folder, ApiError> {
        let api_url = self
            .get_api_1_3_uri()?
            .join("folders")
            .context(ApiErrorKind::Url)?;

        let input = FolderName { name: name.into() };

        let response = client
            .post(api_url)
            .header(AUTHORIZATION, self.auth.clone())
            .json(&input)
            .send()
            .await
            .context(ApiErrorKind::Http)?;

        let response = Self::parse_error(response, StatusCode::OK).await?;
        let response: FoldersResponse =
            serde_json::from_str(&response).context(ApiErrorKind::Json)?;
        let FoldersResponse { folders } = response;

        let folder = match folders.into_iter().nth(0) {
            Some(folder) => folder,
            None => return Err(ApiErrorKind::Unknown.into()),
        };

        Ok(folder)
    }

    pub async fn delete_folder(
        &self,
        client: &Client,
        folder_id: FolderID,
    ) -> Result<(), ApiError> {
        let api_url = self
            .get_api_1_3_uri()?
            .join(&format!("folders/{}", folder_id))
            .context(ApiErrorKind::Url)?;

        let response = client
            .delete(api_url)
            .header(AUTHORIZATION, self.auth.clone())
            .send()
            .await
            .context(ApiErrorKind::Http)?;

        let _response = Self::parse_error(response, StatusCode::OK).await?;
        Ok(())
    }

    pub async fn rename_folder(
        &self,
        client: &Client,
        folder_id: FolderID,
        new_name: &str,
    ) -> Result<(), ApiError> {
        let api_url = self
            .get_api_1_3_uri()?
            .join(&format!("folders/{}", folder_id))
            .context(ApiErrorKind::Url)?;

        let input = FolderName {
            name: new_name.into(),
        };

        let response = client
            .put(api_url)
            .header(AUTHORIZATION, self.auth.clone())
            .json(&input)
            .send()
            .await
            .context(ApiErrorKind::Http)?;

        let _response = Self::parse_error(response, StatusCode::OK).await?;
        Ok(())
    }

    pub async fn mark_folder(&self, client: &Client, folder_id: FolderID, newest_item_id: ItemID) -> Result<(), ApiError> {
        let api_url = self
            .get_api_1_3_uri()?
            .join(&format!("folders/{}/read", folder_id))
            .context(ApiErrorKind::Url)?;

        let input = ItemMarkAllInput { newest_item_id };

        let response = client
            .post(api_url)
            .header(AUTHORIZATION, self.auth.clone())
            .json(&input)
            .send()
            .await
            .context(ApiErrorKind::Http)?;

        let _response = Self::parse_error(response, StatusCode::OK).await?;
        Ok(())
    }

    pub async fn get_feeds(&self, client: &Client) -> Result<Vec<Feed>, ApiError> {
        let api_url = self
            .get_api_1_3_uri()?
            .join("feeds")
            .context(ApiErrorKind::Url)?;

        let response = client
            .get(api_url)
            .header(AUTHORIZATION, self.auth.clone())
            .send()
            .await
            .context(ApiErrorKind::Http)?;

        let response = Self::parse_error(response, StatusCode::OK).await?;
        let response: FeedsResponse =
            serde_json::from_str(&response).context(ApiErrorKind::Json)?;
        let FeedsResponse { feeds, .. } = response;
        Ok(feeds)
    }

    pub async fn create_feed(
        &self,
        client: &Client,
        url: &str,
        folder_id: Option<FolderID>,
    ) -> Result<Feed, ApiError> {
        let api_url = self
            .get_api_1_3_uri()?
            .join("feeds")
            .context(ApiErrorKind::Url)?;

        let input = FeedCreateInput {
            url: url.into(),
            folder_id,
        };

        let response = client
            .post(api_url)
            .header(AUTHORIZATION, self.auth.clone())
            .json(&input)
            .send()
            .await
            .context(ApiErrorKind::Http)?;

        let response = Self::parse_error(response, StatusCode::OK).await?;
        let response: FeedsResponse =
            serde_json::from_str(&response).context(ApiErrorKind::Json)?;
        let FeedsResponse { feeds, .. } = response;

        let feed = match feeds.into_iter().nth(0) {
            Some(feed) => feed,
            None => return Err(ApiErrorKind::Unknown.into()),
        };

        Ok(feed)
    }

    pub async fn delete_feed(&self, client: &Client, feed_id: FeedID) -> Result<(), ApiError> {
        let api_url = self
            .get_api_1_3_uri()?
            .join(&format!("feeds/{}", feed_id))
            .context(ApiErrorKind::Url)?;

        let response = client
            .delete(api_url)
            .header(AUTHORIZATION, self.auth.clone())
            .send()
            .await
            .context(ApiErrorKind::Http)?;

        let _response = Self::parse_error(response, StatusCode::OK).await?;
        Ok(())
    }

    pub async fn move_feed(
        &self,
        client: &Client,
        feed_id: FeedID,
        folder_id: Option<FolderID>,
    ) -> Result<(), ApiError> {
        let api_url = self
            .get_api_1_3_uri()?
            .join(&format!("feeds/{}/move", feed_id))
            .context(ApiErrorKind::Url)?;

        let input = FeedMoveInput { folder_id };

        let response = client
            .post(api_url)
            .header(AUTHORIZATION, self.auth.clone())
            .json(&input)
            .send()
            .await
            .context(ApiErrorKind::Http)?;

        let _response = Self::parse_error(response, StatusCode::OK).await?;
        Ok(())
    }

    pub async fn rename_feed(
        &self,
        client: &Client,
        feed_id: FeedID,
        new_name: &str,
    ) -> Result<(), ApiError> {
        let api_url = self
            .get_api_1_3_uri()?
            .join(&format!("feeds/{}/rename", feed_id))
            .context(ApiErrorKind::Url)?;

        let input = FeedRenameInput {
            feed_title: new_name.into(),
        };

        let response = client
            .post(api_url)
            .header(AUTHORIZATION, self.auth.clone())
            .json(&input)
            .send()
            .await
            .context(ApiErrorKind::Http)?;

        let _response = Self::parse_error(response, StatusCode::OK).await?;
        Ok(())
    }

    pub async fn mark_feed(&self, client: &Client, feed_id: FeedID, newest_item_id: ItemID) -> Result<(), ApiError> {
        let api_url = self
            .get_api_1_3_uri()?
            .join(&format!("feeds/{}/read", feed_id))
            .context(ApiErrorKind::Url)?;

        let input = ItemMarkAllInput { newest_item_id };

        let response = client
            .post(api_url)
            .header(AUTHORIZATION, self.auth.clone())
            .json(&input)
            .send()
            .await
            .context(ApiErrorKind::Http)?;

        let _response = Self::parse_error(response, StatusCode::OK).await?;
        Ok(())
    }

    pub async fn get_items(
        &self,
        client: &Client,
        batch_size: i64,
        offset: Option<u64>,
        item_type: Option<ItemType>,
        id: Option<u64>,
        get_read: Option<bool>,
        oldest_first: Option<bool>,
    ) -> Result<Vec<Item>, ApiError> {
        let api_url = self
            .get_api_1_3_uri()?
            .join("items")
            .context(ApiErrorKind::Url)?;

        let input = ItemQueryInput {
            batch_size,
            offset,
            item_type,
            id,
            get_read,
            oldest_first,
        };

        let response = client
            .get(api_url)
            .header(AUTHORIZATION, self.auth.clone())
            .json(&input)
            .send()
            .await
            .context(ApiErrorKind::Http)?;

        let response = Self::parse_error(response, StatusCode::OK).await?;
        let response: ItemsResponse =
            serde_json::from_str(&response).context(ApiErrorKind::Json)?;
        let ItemsResponse { items, .. } = response;
        Ok(items)
    }

    pub async fn get_updated_items(
        &self,
        client: &Client,
        last_modified: u64,
        item_type: Option<ItemType>,
        id: Option<u64>,
    ) -> Result<Vec<Item>, ApiError> {
        let api_url = self
            .get_api_1_3_uri()?
            .join("items/updated")
            .context(ApiErrorKind::Url)?;

        let input = ItemUpdatedQueryInput {
            last_modified,
            item_type,
            id,
        };

        let response = client
            .get(api_url)
            .header(AUTHORIZATION, self.auth.clone())
            .json(&input)
            .send()
            .await
            .context(ApiErrorKind::Http)?;

        let response = Self::parse_error(response, StatusCode::OK).await?;
        let response: ItemsResponse =
            serde_json::from_str(&response).context(ApiErrorKind::Json)?;
        let ItemsResponse { items, .. } = response;
        Ok(items)
    }

    pub async fn mark_item_read(&self, client: &Client, item_id: ItemID) -> Result<(), ApiError> {
        let api_url = self
            .get_api_1_3_uri()?
            .join(&format!("items/{}/read", item_id))
            .context(ApiErrorKind::Url)?;

        let response = client
            .post(api_url)
            .header(AUTHORIZATION, self.auth.clone())
            .send()
            .await
            .context(ApiErrorKind::Http)?;

        let _response = Self::parse_error(response, StatusCode::OK).await?;
        Ok(())
    }

    pub async fn mark_items_read(
        &self,
        client: &Client,
        item_ids: Vec<ItemID>,
    ) -> Result<(), ApiError> {
        let api_url = self
            .get_api_1_3_uri()?
            .join("items/read/multiple")
            .context(ApiErrorKind::Url)?;

        let input = ItemMarkInput { item_ids };

        let response = client
            .post(api_url)
            .header(AUTHORIZATION, self.auth.clone())
            .json(&input)
            .send()
            .await
            .context(ApiErrorKind::Http)?;

        let _response = Self::parse_error(response, StatusCode::OK).await?;
        Ok(())
    }

    pub async fn mark_item_unread(&self, client: &Client, item_id: ItemID) -> Result<(), ApiError> {
        let api_url = self
            .get_api_1_3_uri()?
            .join(&format!("items/{}/unread", item_id))
            .context(ApiErrorKind::Url)?;

        let response = client
            .post(api_url)
            .header(AUTHORIZATION, self.auth.clone())
            .send()
            .await
            .context(ApiErrorKind::Http)?;

        let _response = Self::parse_error(response, StatusCode::OK).await?;
        Ok(())
    }

    pub async fn mark_items_unread(
        &self,
        client: &Client,
        item_ids: Vec<ItemID>,
    ) -> Result<(), ApiError> {
        let api_url = self
            .get_api_1_3_uri()?
            .join("items/unread/multiple")
            .context(ApiErrorKind::Url)?;

        let input = ItemMarkInput { item_ids };

        let response = client
            .post(api_url)
            .header(AUTHORIZATION, self.auth.clone())
            .json(&input)
            .send()
            .await
            .context(ApiErrorKind::Http)?;

        let _response = Self::parse_error(response, StatusCode::OK).await?;
        Ok(())
    }

    pub async fn mark_item_starred(
        &self,
        client: &Client,
        item_id: ItemID,
    ) -> Result<(), ApiError> {
        let api_url = self
            .get_api_1_3_uri()?
            .join(&format!("items/{}/star", item_id))
            .context(ApiErrorKind::Url)?;

        let response = client
            .post(api_url)
            .header(AUTHORIZATION, self.auth.clone())
            .send()
            .await
            .context(ApiErrorKind::Http)?;

        let _response = Self::parse_error(response, StatusCode::OK).await?;
        Ok(())
    }

    pub async fn mark_items_starred(
        &self,
        client: &Client,
        item_ids: Vec<ItemID>,
    ) -> Result<(), ApiError> {
        let api_url = self
            .get_api_1_3_uri()?
            .join("items/star/multiple")
            .context(ApiErrorKind::Url)?;

        let input = ItemMarkInput { item_ids };

        let response = client
            .post(api_url)
            .header(AUTHORIZATION, self.auth.clone())
            .json(&input)
            .send()
            .await
            .context(ApiErrorKind::Http)?;

        let _response = Self::parse_error(response, StatusCode::OK).await?;
        Ok(())
    }

    pub async fn mark_item_unstarred(
        &self,
        client: &Client,
        item_id: ItemID,
    ) -> Result<(), ApiError> {
        let api_url = self
            .get_api_1_3_uri()?
            .join(&format!("items/{}/unstar", item_id))
            .context(ApiErrorKind::Url)?;

        let response = client
            .post(api_url)
            .header(AUTHORIZATION, self.auth.clone())
            .send()
            .await
            .context(ApiErrorKind::Http)?;

        let _response = Self::parse_error(response, StatusCode::OK).await?;
        Ok(())
    }

    pub async fn mark_items_unstarred(
        &self,
        client: &Client,
        item_ids: Vec<ItemID>,
    ) -> Result<(), ApiError> {
        let api_url = self
            .get_api_1_3_uri()?
            .join("items/unstar/multiple")
            .context(ApiErrorKind::Url)?;

        let input = ItemMarkInput { item_ids };

        let response = client
            .post(api_url)
            .header(AUTHORIZATION, self.auth.clone())
            .json(&input)
            .send()
            .await
            .context(ApiErrorKind::Http)?;

        let _response = Self::parse_error(response, StatusCode::OK).await?;
        Ok(())
    }

    pub async fn mark_all_items_read(
        &self,
        client: &Client,
        newest_item_id: ItemID,
    ) -> Result<(), ApiError> {
        let api_url = self
            .get_api_1_3_uri()?
            .join("items/read")
            .context(ApiErrorKind::Url)?;

        let input = ItemMarkAllInput { newest_item_id };

        let response = client
            .post(api_url)
            .header(AUTHORIZATION, self.auth.clone())
            .json(&input)
            .send()
            .await
            .context(ApiErrorKind::Http)?;

        let _response = Self::parse_error(response, StatusCode::OK).await?;
        Ok(())
    }
}
